@@ -11,6 +11,11 @@ from tuxmake.toolchain import Toolchain
 from tuxmake.build import build
 from tuxmake.build import Build
 from tuxmake.build import defaults
+from tuxmake.build import Terminated
+from tuxmake.build import get_image
+from tuxmake.build import DEFAULT_CONTAINER_REGISTRY
+from tuxmake.target import Command
+from tuxmake.target import default_compression
 import tuxmake.exceptions
 
 
@@ -30,21 +35,11 @@ def check_artifacts(mocker):
     return mocker.patch("tuxmake.build.Build.check_artifacts", return_value=True)
 
 
-@pytest.fixture()
-def Popen(mocker, check_artifacts):
-    _Popen = mocker.patch("subprocess.Popen")
-    _Popen.return_value.communicate.return_value = (
-        mocker.MagicMock(),
-        mocker.MagicMock(),
-    )
-    return _Popen
-
-
 # Disable the metadata extraction for non-metadata related tests since its
 # pretty slow.
 @pytest.fixture(autouse=True)
-def disable_metadata(mocker):
-    return mocker.patch("tuxmake.build.Build.extract_metadata")
+def collect_metadata(mocker):
+    return mocker.patch("tuxmake.build.Build.collect_metadata")
 
 
 def args(called):
@@ -55,180 +50,197 @@ def kwargs(called):
     return called.call_args[1]
 
 
-def test_invalid_directory(tmp_path):
-    (tmp_path / "Makefile").touch()
-    with pytest.raises(tuxmake.exceptions.UnrecognizedSourceTree):
-        build(tree=tmp_path)
+class TestBasicFunctionality:
+    def test_invalid_directory(self, tmp_path):
+        (tmp_path / "Makefile").touch()
+        with pytest.raises(tuxmake.exceptions.UnrecognizedSourceTree):
+            build(tree=tmp_path)
+
+    def test_build(self, linux, home, kernel):
+        result = build(tree=linux)
+        assert kernel in result.artifacts["kernel"]
+        assert (home / ".cache/tuxmake/builds/1" / kernel).exists()
+        assert result.passed
+
+    def test_build_with_output_dir(self, linux, output_dir, kernel):
+        result = build(tree=linux, output_dir=output_dir)
+        assert kernel in result.artifacts["kernel"]
+        assert (output_dir / kernel).exists()
+        assert result.output_dir == output_dir
+
+    def test_build_with_build_dir(self, linux, tmp_path):
+        build(tree=linux, build_dir=tmp_path)
+        assert (tmp_path / ".config").exists
+
+    def test_no_directory_created_unecessarily(self, linux, home):
+        Build(tree=linux)
+        assert len(list(home.glob("*"))) == 0
+
+    def test_no_directory_created_unecessarily_with_explicit_paths(
+        self, linux, tmp_path
+    ):
+        Build(tree=linux, output_dir=tmp_path / "output", build_dir=tmp_path / "build")
+        assert not (tmp_path / "output").exists()
+        assert not (tmp_path / "build").exists()
+
+    def test_unsupported_target(self, linux):
+        with pytest.raises(tuxmake.exceptions.UnsupportedTarget):
+            build(tree=linux, targets=["unknown-target"])
 
 
-def test_build(linux, home, kernel):
-    result = build(tree=linux)
-    assert kernel in result.artifacts["kernel"]
-    assert (home / ".cache/tuxmake/builds/1" / kernel).exists()
-    assert result.passed
+class TestKconfig:
+    def test_kconfig_default(self, linux, Popen):
+        b = Build(tree=linux, targets=["config"])
+        b.build(b.targets[0])
+        assert "defconfig" in args(Popen)
 
+    def test_kconfig_named(self, linux, Popen):
+        b = Build(tree=linux, targets=["config"], kconfig="fooconfig")
+        b.build(b.targets[0])
+        assert "fooconfig" in args(Popen)
 
-def test_build_with_output_dir(linux, output_dir, kernel):
-    result = build(tree=linux, output_dir=output_dir)
-    assert kernel in result.artifacts["kernel"]
-    assert (output_dir / kernel).exists()
-    assert result.output_dir == output_dir
+    def test_kconfig_named_invalid(self, linux, mocker):
+        with pytest.raises(tuxmake.exceptions.UnsupportedKconfig):
+            build(tree=linux, targets=["config"], kconfig="foobar")
 
+    def test_kconfig_url(self, linux, mocker, output_dir):
+        response = mocker.MagicMock()
+        response.getcode.return_value = 200
+        response.read.return_value = b"CONFIG_FOO=y\nCONFIG_BAR=y\n"
+        mocker.patch("urllib.request.urlopen", return_value=response)
 
-def test_build_with_build_dir(linux, tmp_path):
-    build(tree=linux, build_dir=tmp_path)
-    assert (tmp_path / ".config").exists
+        build(
+            tree=linux,
+            targets=["config"],
+            kconfig="https://example.com/config.txt",
+            output_dir=output_dir,
+        )
+        config = output_dir / "config"
+        assert "CONFIG_FOO=y\nCONFIG_BAR=y\n" in config.read_text()
 
+    def test_kconfig_url_not_found(self, linux, mocker):
+        mocker.patch(
+            "urllib.request.urlopen",
+            side_effect=urllib.error.HTTPError(
+                "https://example.com/config.txt", 404, "Not Found", {}, None
+            ),
+        )
 
-def test_no_directory_created_unecessarily(linux, home):
-    Build(tree=linux)
-    assert len(list(home.glob("*"))) == 0
+        with pytest.raises(tuxmake.exceptions.InvalidKConfig):
+            build(
+                tree=linux, targets=["config"], kconfig="https://example.com/config.txt"
+            )
 
+    def test_kconfig_localfile(self, linux, tmp_path, output_dir):
+        extra_config = tmp_path / "extra_config"
+        extra_config.write_text("CONFIG_XYZ=y\nCONFIG_ABC=m\n")
+        build(
+            tree=linux,
+            targets=["config"],
+            kconfig=str(extra_config),
+            output_dir=output_dir,
+        )
+        config = output_dir / "config"
+        assert "CONFIG_XYZ=y\nCONFIG_ABC=m\n" in config.read_text()
 
-def test_no_directory_created_unecessarily_with_explicit_paths(linux, tmp_path):
-    Build(tree=linux, output_dir=tmp_path / "output", build_dir=tmp_path / "build")
-    assert not (tmp_path / "output").exists()
-    assert not (tmp_path / "build").exists()
+    def test_kconfig_add_url(self, linux, mocker, output_dir):
+        response = mocker.MagicMock()
+        response.getcode.return_value = 200
+        response.read.return_value = b"CONFIG_FOO=y\nCONFIG_BAR=y\n"
+        mocker.patch("urllib.request.urlopen", return_value=response)
 
+        build(
+            tree=linux,
+            targets=["config"],
+            kconfig="defconfig",
+            kconfig_add=["https://example.com/config.txt"],
+            output_dir=output_dir,
+        )
+        config = output_dir / "config"
+        assert "CONFIG_FOO=y\nCONFIG_BAR=y\n" in config.read_text()
 
-def test_unsupported_target(linux):
-    with pytest.raises(tuxmake.exceptions.UnsupportedTarget):
-        build(tree=linux, targets=["unknown-target"])
+    def test_kconfig_add_localfile(self, linux, tmp_path, output_dir):
+        extra_config = tmp_path / "extra_config"
+        extra_config.write_text("CONFIG_XYZ=y\nCONFIG_ABC=m\n")
+        build(
+            tree=linux,
+            targets=["config"],
+            kconfig_add=[str(extra_config)],
+            output_dir=output_dir,
+        )
+        config = output_dir / "config"
+        assert "CONFIG_XYZ=y\nCONFIG_ABC=m\n" in config.read_text()
 
+    def test_kconfig_add_inline(self, linux, output_dir):
+        build(
+            tree=linux,
+            targets=["config"],
+            kconfig_add=["CONFIG_FOO=y"],
+            output_dir=output_dir,
+        )
+        config = output_dir / "config"
+        assert "CONFIG_FOO=y\n" in config.read_text()
 
-def test_kconfig_default(linux, Popen):
-    b = Build(tree=linux, targets=["config"])
-    b.build(b.targets[0])
-    assert "defconfig" in args(Popen)
+    def test_kconfig_add_inline_not_set(self, linux, output_dir):
+        build(
+            tree=linux,
+            targets=["config"],
+            kconfig_add=["# CONFIG_FOO is not set"],
+            output_dir=output_dir,
+        )
+        config = output_dir / "config"
+        assert "CONFIG_FOO is not set\n" in config.read_text()
 
+    def test_kconfig_add_inline_set_to_no(self, linux, output_dir):
+        build(
+            tree=linux,
+            targets=["config"],
+            kconfig_add=["CONFIG_FOO=n"],
+            output_dir=output_dir,
+        )
+        config = output_dir / "config"
+        assert "CONFIG_FOO=n\n" in config.read_text()
 
-def test_kconfig_named(linux, Popen):
-    b = Build(tree=linux, targets=["config"], kconfig="fooconfig")
-    b.build(b.targets[0])
-    assert "fooconfig" in args(Popen)
+    def test_kconfig_add_in_tree(self, linux, output_dir):
+        build(
+            tree=linux,
+            targets=["config"],
+            kconfig_add=["kvm_guest.config", "qemu-gdb.config"],
+            output_dir=output_dir,
+        )
+        config = output_dir / "config"
+        assert ("CONFIG_KVM_GUEST=y") in config.read_text()
+        assert ("CONFIG_DEBUG_INFO=y") in config.read_text()
 
+    def test_kconfig_add_explicit_make_target(self, linux, output_dir):
+        build(
+            tree=linux,
+            targets=["config"],
+            kconfig_add=["make:kselftest-merge"],
+            output_dir=output_dir,
+        )
+        config = output_dir / "config"
+        assert "CONFIG_KSELFTEST_MERGE=y" in config.read_text()
 
-def test_kconfig_named_invalid(linux, mocker):
-    with pytest.raises(tuxmake.exceptions.UnsupportedKconfig):
-        build(tree=linux, targets=["config"], kconfig="foobar")
+    def test_kconfig_add_explicit_interactive_make_target(self, linux, Popen):
+        build = Build(
+            tree=linux,
+            targets=["config"],
+            kconfig_add=["imake:menuconfig"],
+        )
+        build.build(build.targets[0])
+        cmd = build.targets[-1].commands[1]
+        assert cmd == ["{make}", "menuconfig"]
+        assert cmd.interactive
 
+    def test_kconfig_add_invalid(self, linux):
+        with pytest.raises(tuxmake.exceptions.UnsupportedKconfigFragment):
+            build(tree=linux, targets=["config"], kconfig_add=["foo"])
 
-def test_kconfig_url(linux, mocker, output_dir):
-    response = mocker.MagicMock()
-    response.getcode.return_value = 200
-    response.read.return_value = b"CONFIG_FOO=y\nCONFIG_BAR=y\n"
-    mocker.patch("urllib.request.urlopen", return_value=response)
-
-    build(
-        tree=linux,
-        targets=["config"],
-        kconfig="https://example.com/config.txt",
-        output_dir=output_dir,
-    )
-    config = output_dir / "config"
-    assert "CONFIG_FOO=y\nCONFIG_BAR=y\n" in config.read_text()
-
-
-def test_kconfig_url_not_found(linux, mocker):
-    mocker.patch(
-        "urllib.request.urlopen",
-        side_effect=urllib.error.HTTPError(
-            "https://example.com/config.txt", 404, "Not Found", {}, None
-        ),
-    )
-
-    with pytest.raises(tuxmake.exceptions.InvalidKConfig):
-        build(tree=linux, targets=["config"], kconfig="https://example.com/config.txt")
-
-
-def test_kconfig_localfile(linux, tmp_path, output_dir):
-    extra_config = tmp_path / "extra_config"
-    extra_config.write_text("CONFIG_XYZ=y\nCONFIG_ABC=m\n")
-    build(
-        tree=linux, targets=["config"], kconfig=str(extra_config), output_dir=output_dir
-    )
-    config = output_dir / "config"
-    assert "CONFIG_XYZ=y\nCONFIG_ABC=m\n" in config.read_text()
-
-
-def test_kconfig_add_url(linux, mocker, output_dir):
-    response = mocker.MagicMock()
-    response.getcode.return_value = 200
-    response.read.return_value = b"CONFIG_FOO=y\nCONFIG_BAR=y\n"
-    mocker.patch("urllib.request.urlopen", return_value=response)
-
-    build(
-        tree=linux,
-        targets=["config"],
-        kconfig="defconfig",
-        kconfig_add=["https://example.com/config.txt"],
-        output_dir=output_dir,
-    )
-    config = output_dir / "config"
-    assert "CONFIG_FOO=y\nCONFIG_BAR=y\n" in config.read_text()
-
-
-def test_kconfig_add_localfile(linux, tmp_path, output_dir):
-    extra_config = tmp_path / "extra_config"
-    extra_config.write_text("CONFIG_XYZ=y\nCONFIG_ABC=m\n")
-    build(
-        tree=linux,
-        targets=["config"],
-        kconfig_add=[str(extra_config)],
-        output_dir=output_dir,
-    )
-    config = output_dir / "config"
-    assert "CONFIG_XYZ=y\nCONFIG_ABC=m\n" in config.read_text()
-
-
-def test_kconfig_add_inline(linux, output_dir):
-    build(
-        tree=linux,
-        targets=["config"],
-        kconfig_add=["CONFIG_FOO=y"],
-        output_dir=output_dir,
-    )
-    config = output_dir / "config"
-    assert "CONFIG_FOO=y\n" in config.read_text()
-
-
-def test_kconfig_add_inline_not_set(linux, output_dir):
-    build(
-        tree=linux,
-        targets=["config"],
-        kconfig_add=["# CONFIG_FOO is not set"],
-        output_dir=output_dir,
-    )
-    config = output_dir / "config"
-    assert "CONFIG_FOO is not set\n" in config.read_text()
-
-
-def test_kconfig_add_inline_set_to_no(linux, output_dir):
-    build(
-        tree=linux,
-        targets=["config"],
-        kconfig_add=["CONFIG_FOO=n"],
-        output_dir=output_dir,
-    )
-    config = output_dir / "config"
-    assert "CONFIG_FOO=n\n" in config.read_text()
-
-
-def test_kconfig_add_in_tree(linux, output_dir):
-    build(
-        tree=linux,
-        targets=["config"],
-        kconfig_add=["kvm_guest.config", "qemu-gdb.config"],
-        output_dir=output_dir,
-    )
-    config = output_dir / "config"
-    assert ("CONFIG_KVM_GUEST=y") in config.read_text()
-    assert ("CONFIG_DEBUG_INFO=y") in config.read_text()
-
-
-def test_kconfig_add_invalid(linux):
-    with pytest.raises(tuxmake.exceptions.UnsupportedKconfigFragment):
-        build(tree=linux, targets=["config"], kconfig_add=["foo"])
+    def test_kconfig_from_source_tree(self, linux):
+        b = build(tree=linux, targets=["config"], kconfig="config/test.config")
+        config = b.output_dir / "config"
+        assert "CONFIG_IN_SOURCE_CONFIG_FILE=y" in config.read_text()
 
 
 def test_output_dir(linux, output_dir, kernel):
@@ -239,6 +251,13 @@ def test_output_dir(linux, output_dir, kernel):
     assert "arch" not in artifacts
 
 
+def test_copies_artifacts_from_failed_targets(linux, output_dir, mocker):
+    mocker.patch("tuxmake.build.Build.check_artifacts", return_value=False)
+    build(tree=linux, output_dir=output_dir, targets=["config"])
+    artifacts = [str(f.name) for f in output_dir.glob("*")]
+    assert "config" in artifacts
+
+
 def test_saves_log(linux):
     result = build(tree=linux)
     artifacts = [str(f.name) for f in result.output_dir.glob("*")]
@@ -248,6 +267,12 @@ def test_saves_log(linux):
     assert "make --silent" in log.read_text()
 
 
+def test_timestamp_in_debug_log(linux):
+    result = build(tree=linux)
+    log = result.output_dir / "build-debug.log"
+    assert "00:00" in log.read_text()
+
+
 def test_build_failure(linux, kernel, monkeypatch):
     monkeypatch.setenv("FAIL", "kernel")
     result = build(tree=linux, targets=["config", "kernel"])
@@ -271,6 +296,25 @@ def test_concurrency_set(linux, Popen):
     assert "--jobs=99" in args(Popen)
 
 
+def test_fail_fast(linux, mocker, Popen):
+    b = Build(tree=linux, targets=["config"], fail_fast=True)
+    b.build(b.targets[0])
+    assert "--keep-going" not in args(Popen)
+
+
+def test_fail_fast_aborts_build(linux, monkeypatch):
+    """
+    `dtbs` do not depend on `kernel`; normally, `dtbs` will still be built even
+    if the kernel build fails. When fail_fast is set, though, the build stops
+    after the first target fails, regardless of dependencies.
+    """
+    b = Build(tree=linux, fail_fast=True, target_arch="arm64")
+    monkeypatch.setenv("FAIL", "kernel")
+    b.run()
+    assert b.status["default"].failed
+    assert b.status["dtbs"].skipped
+
+
 def test_verbose(linux, mocker, Popen):
     b = Build(tree=linux, targets=["config"], verbose=True)
     b.build(b.targets[0])
@@ -289,40 +333,78 @@ def test_quiet(linux, capfd):
     assert "I:" not in err
 
 
-def test_ctrl_c(linux, mocker, Popen):
-    mocker.patch("tuxmake.build.Build.logger")
-    process = mocker.MagicMock()
-    Popen.return_value = process
-    process.communicate.side_effect = KeyboardInterrupt()
-    with pytest.raises(SystemExit):
+class TestInterruptedBuild:
+    @pytest.fixture
+    def interrupted(self, mocker, Popen):
+        process = mocker.MagicMock()
+        Popen.return_value = process
+        process.communicate.side_effect = KeyboardInterrupt()
+        return process
+
+    def test_ctrl_c(self, linux, interrupted):
         b = Build(tree=linux)
-        b.build(b.targets[0])
-    process.terminate.assert_called()
+        res = b.build(b.targets[0])
+        interrupted.terminate.assert_called()
+        assert res.failed
+        assert b.interrupted
 
+    def test_ctrl_c_skips_all_other_targets(self, linux, interrupted, mocker):
+        b = Build(tree=linux)
+        real_build = b.build
+        mock_build = mocker.patch("tuxmake.build.Build.build", wraps=real_build)
+        b.build_all_targets()
+        expected_statuses = ["FAIL"] + ["SKIP" for _ in b.targets[1:]]
+        statuses = [b.status[t.name].status for t in b.targets]
+        assert statuses == expected_statuses
+        assert mock_build.call_count == 1
 
-def test_always_run_cleanup(linux, mocker):
-    build = Build(tree=linux)
-    mocker.patch(
-        "tuxmake.build.Build.build_all_targets", side_effect=KeyboardInterrupt()
-    )
-    with pytest.raises(KeyboardInterrupt):
-        build.run()
-    assert not build.build_dir.exists()
+    def test_always_run_cleanup(self, linux, mocker):
+        build = Build(tree=linux)
+        mocker.patch(
+            "tuxmake.build.Build.build_all_targets", side_effect=KeyboardInterrupt()
+        )
+        with pytest.raises(KeyboardInterrupt):
+            build.run()
+        assert not build.build_dir.exists()
 
+    def test_cleans_up_even_if_prepare_fails(self, linux, mocker):
+        build = Build(tree=linux)
+        mocker.patch("tuxmake.build.Build.prepare", side_effect=KeyboardInterrupt())
+        with pytest.raises(KeyboardInterrupt):
+            build.run()
+        assert not build.build_dir.exists()
 
-def test_cleans_up_even_if_prepare_fails(linux, mocker):
-    build = Build(tree=linux)
-    mocker.patch("tuxmake.build.Build.prepare", side_effect=KeyboardInterrupt())
-    with pytest.raises(KeyboardInterrupt):
-        build.run()
-    assert not build.build_dir.exists()
+    def test_copies_artifacts_even_when_interrupted(self, linux, mocker):
+        build = Build(tree=linux)
+        mocker.patch(
+            "tuxmake.build.Build.build_all_targets", side_effect=KeyboardInterrupt()
+        )
+        copy_artifacts = mocker.patch("tuxmake.build.Build.copy_artifacts")
+        with pytest.raises(KeyboardInterrupt):
+            build.run()
+        assert copy_artifacts.call_count > 0
 
+    def test_gets_metadata_even_when_interrupted(self, linux, mocker, collect_metadata):
+        build = Build(tree=linux)
+        mocker.patch(
+            "tuxmake.build.Build.build_all_targets", side_effect=KeyboardInterrupt()
+        )
+        with pytest.raises(KeyboardInterrupt):
+            build.run()
+        assert collect_metadata.call_count == 1
+        assert (build.output_dir / "metadata.json").exists()
 
-def test_existing_build_dir(linux, home):
-    (home / ".cache" / "tuxmake" / "builds" / "current").mkdir(parents=True)
-    build = Build(tree=linux)
-    with pytest.raises(tuxmake.exceptions.TuxMakeException):
-        build.build_dir
+    def test_does_not_collect_metadata_when_runtime_preparation_fails(
+        self, linux, mocker, collect_metadata
+    ):
+        build = Build(tree=linux)
+        mocker.patch(
+            "tuxmake.runtime.NullRuntime.prepare",
+            side_effect=RuntimeError("PREPARE FAILED"),
+        )
+        with pytest.raises(RuntimeError):
+            build.run()
+        assert collect_metadata.call_count == 0
 
 
 clang_version = int(
@@ -421,6 +503,13 @@ class TestToolchain:
         with pytest.raises(tuxmake.exceptions.UnsupportedToolchain):
             Toolchain("foocc")
 
+    def test_prepare_warns_about_versioned_toolchain(self, linux, mocker):
+        build = Build(tree=linux, toolchain="gcc-10", runtime="null")
+        log = mocker.patch("tuxmake.build.Build.log")
+        build.prepare()
+        log.assert_called()
+        assert "versioned toolchains" in log.call_args[0][0]
+
 
 class TestDebugKernel:
     def test_build_with_debugkernel(self, linux):
@@ -437,6 +526,11 @@ class TestDebugKernel:
         assert "vmlinux.xz" in artifacts
         assert "System.map" in artifacts
 
+    def test_reuse_build_directory(self, linux, tmp_path):
+        build(tree=linux, targets=["config", "debugkernel"], build_dir=tmp_path)
+        r = build(tree=linux, targets=["config", "debugkernel"], build_dir=tmp_path)
+        assert r.passed
+
 
 class TestRunCmd:
     def test_pass(self, linux):
@@ -473,6 +567,21 @@ class TestModules:
         artifacts = [str(f.name) for f in result.output_dir.glob("*")]
         assert "modules.tar.xz" not in artifacts
 
+    def test_cleans_up_modules_install_directory(self, tmp_path, linux_rw):
+        build_dir = tmp_path / "build"
+        build(tree=linux_rw, build_dir=build_dir, targets=["modules"])
+        makefile = str(linux_rw / "Makefile")
+        subprocess.check_call(
+            ["sed", "-i", "-e", "s/^VERSION.*/VERSION = 9.9.9.9.9/", makefile]
+        )
+        b = build(tree=linux_rw, build_dir=build_dir, targets=["modules"])
+        tarball = b.output_dir / "modules.tar.xz"
+        items = subprocess.check_output(
+            ["tar", "taf", str(tarball)], encoding="utf-8"
+        ).split("\n")
+        modules = [i for i in items if i.endswith("/ext4.ko")]
+        assert modules == ["lib/modules/9.9.9.9.9/kernel/fs/ext4/ext4.ko"]
+
 
 def tarball_contents(tarball):
     return subprocess.check_output(["tar", "taf", tarball]).decode("utf-8").splitlines()
@@ -514,6 +623,8 @@ class TestDtbsLegacy:
         subprocess.check_call(
             ["sed", "-i", "-e", "s/dtbs_install/XXXX/g", str(linux_rw / "Makefile")]
         )
+        with (linux_rw / "arch/arm64/Makefile").open("a") as f:
+            f.write("\n\ndtbs:\n\t@echo build dtbs\n")
         return linux_rw
 
     def test_collect_dtbs_manually_without_dtbs_install(self, oldlinux):
@@ -532,7 +643,7 @@ class TestDtbsLegacy:
     ):
         build = Build(tree=oldlinux, target_arch="arm64")
         dtbs_legacy = [t for t in build.targets if t.name == "dtbs-legacy"][0]
-        monkeypatch.setattr(dtbs_legacy, "commands", [["false"]])
+        monkeypatch.setattr(dtbs_legacy, "commands", [Command(["false"])])
         build.run()
         assert build.failed
 
@@ -550,9 +661,9 @@ class TestTargetDependencies:
         assert result.status["kernel"].passed
 
     def test_recursive_dependencies(self, linux):
-        result = build(tree=linux, targets=["modules"])
+        result = build(tree=linux, targets=["kernel"])
+        assert result.status["default"].passed
         assert result.status["config"].passed
-        assert result.status["modules"].passed
 
 
 class TestRuntime:
@@ -565,13 +676,11 @@ class TestRuntime:
         assert build.runtime
 
     def test_interactive_command(self, linux, mocker):
-        runtime = mocker.patch("tuxmake.build.get_runtime").return_value
+        runtime = mocker.patch("tuxmake.runtime.Runtime.get").return_value
         runtime.get_command_line.return_value = ["true"]
         build = Build(tree=linux, runtime="docker")
         build.run_cmd(["true"], interactive=True)
-        runtime.get_command_line.assert_called_with(
-            build, ["true"], True, offline=build.offline
-        )
+        assert kwargs(runtime.run_cmd)["interactive"]
 
 
 class TestEnvironment:
@@ -581,8 +690,8 @@ class TestEnvironment:
             environment={"KCONFIG_ALLCONFIG": "foo.config"},
             targets=["config"],
         )
-        b.build(b.targets[0])
-        assert kwargs(Popen)["env"]["KCONFIG_ALLCONFIG"] == "foo.config"
+        b.prepare()
+        assert b.runtime.environment["KCONFIG_ALLCONFIG"] == "foo.config"
 
 
 class TestMakeVariables:
@@ -599,10 +708,19 @@ class TestMakeVariables:
 class TestCompilerWrappers:
     def test_ccache(self, linux, Popen):
         b = Build(tree=linux, targets=["config"], wrapper="ccache")
-        b.build(b.targets[0])
-        assert "CC=ccache gcc" in args(Popen)
-        assert "HOSTCC=ccache gcc" in args(Popen)
-        assert "CCACHE_DIR" in kwargs(Popen)["env"]
+        b.prepare()
+        assert "CCACHE_DIR" in b.runtime.environment
+        makevars = b.makevars
+        assert makevars["CC"] == "ccache gcc"
+        assert makevars["HOSTCC"] == "ccache gcc"
+
+    def test_ccache_with_container_runtime(self, linux, Popen, mocker):
+        b = Build(tree=linux, targets=["config"], wrapper="ccache", runtime="podman")
+        mocker.patch("tuxmake.runtime.ContainerRuntime.prepare")
+        mocker.patch("tuxmake.wrapper.Wrapper.prepare_runtime")
+        b.prepare()
+        cache_dir = b.wrapper.environment["CCACHE_DIR"]
+        assert (cache_dir, cache_dir) in b.runtime.volumes
 
     def test_ccache_gcc_v(self, linux, Popen):
         b = Build(tree=linux, targets=["config"], toolchain="gcc-10", wrapper="ccache")
@@ -637,6 +755,13 @@ class TestCompilerWrappers:
         b.build(b.targets[0])
         assert "CC=ccache clang" in args(Popen)
 
+    def test_sccache_with_path(self, linux, mocker, Popen):
+        add_volume = mocker.patch("tuxmake.runtime.Runtime.add_volume")
+        b = Build(tree=linux, wrapper="/path/to/sccache")
+        b.prepare()
+        volumes = [call[0] for call in add_volume.call_args_list]
+        assert ("/path/to/sccache", "/usr/local/bin/sccache") in volumes
+
 
 @pytest.mark.skipif(
     [int(n) for n in pytest.__version__.split(".")] < [3, 10], reason="old pytest"
@@ -658,6 +783,11 @@ class TestMetadata:
             is not None
         )
 
+    def test_kernelrelease(self, metadata):
+        assert metadata["source"]["kernelrelease"].startswith(
+            metadata["source"]["kernelversion"]
+        )
+
     def test_metadata_file(self, metadata):
         assert type(metadata) is dict
 
@@ -716,7 +846,9 @@ class TestDebug:
 
     @pytest.fixture
     def debug_build(self, linux):
-        return Build(tree=linux, debug=True, environment={"FOO": "BAR"})
+        b = Build(tree=linux, debug=True, environment={"FOO": "BAR"})
+        b.prepare()
+        return b
 
     @pytest.fixture
     def err(self, debug_build, mocker, capfd):
@@ -747,7 +879,7 @@ class TestPrepare:
         )
         mocker.patch(
             "tuxmake.runtime.NullRuntime.prepare",
-            side_effect=lambda _: order.append("runtime"),
+            side_effect=lambda: order.append("runtime"),
         )
         mocker.patch(
             "tuxmake.wrapper.Wrapper.prepare_runtime",
@@ -821,6 +953,11 @@ class TestKselftest:
         names = [t.name for t in build.targets]
         assert names == ["config", "kselftest"]
 
+    def test_kselftest_merge_runs_right_after_config_and_before_default(self, linux):
+        build = Build(tree=linux, targets=["config", "kernel", "kselftest-merge"])
+        names = [t.name for t in build.targets]
+        assert names == ["config", "kselftest-merge", "default", "kernel"]
+
 
 class TestBPFKselftest:
     def test_bpf_kselftest_kconfig_add(self, linux, output_dir):
@@ -829,6 +966,7 @@ class TestBPFKselftest:
         assert result.kconfig_add == kselftest_target.kconfig_add
         config = output_dir / "config"
         assert ("CONFIG_DEBUG_INFO_BTF=y") in config.read_text()
+        assert ("CONFIG_INCLUDE_IN_TREE_FILE=y") in config.read_text()
 
 
 class TestHeaders:
@@ -839,14 +977,32 @@ class TestHeaders:
 
 
 class TestCheckEnvironment:
-    def test_basics(self, linux, Popen):
-        Popen.return_value.returncode = 0
+    @pytest.fixture
+    def get_command_output(self, mocker):
+        return mocker.patch("tuxmake.build.Build.get_command_output")
+
+    @pytest.fixture
+    def run_cmd(self, mocker):
+        rcmd = mocker.patch("tuxmake.build.Build.run_cmd")
+        rcmd.return_value.returncode = 0
+        return rcmd
+
+    def test_basics(self, linux, get_command_output, run_cmd):
         build = Build(tree=linux, target_arch="arm64")
+        ccc = "arm-linux-gnu-"
+        get_command_output.return_value = ccc
         build.check_environment()
-        cmdline = args(Popen)
+        cmdline = args(run_cmd)
         assert cmdline[0].endswith("/tuxmake-check-environment")
         assert cmdline[1] == "arm64_gcc"
         assert cmdline[2] == "aarch64-linux-gnu-"
+        assert cmdline[3] == ccc
+
+    def test_CROSS_COMPILE_COMPAT_not_found(self, linux, run_cmd, get_command_output):
+        build = Build(tree=linux, target_arch="arm64")
+        get_command_output.return_value = ""
+        build.check_environment()
+        assert args(run_cmd)[3] == ""
 
     def test_fails(self, linux, Popen):
         Popen.return_value.returncode = 1
@@ -866,4 +1022,82 @@ class TestReproducible:
     def test_reproducible_sets_constant_values(self, linux):
         build1 = Build(tree=linux)
         build2 = Build(tree=linux)
-        assert build1.environment == build2.environment
+        ts = "KBUILD_BUILD_TIMESTAMP"
+        assert build1.environment[ts] == build2.environment[ts]
+
+
+class TestTerminated:
+    def test_signal_handler_raises_exception(self):
+        with pytest.raises(Terminated):
+            Terminated.handle_signal(15, None)
+
+
+def q(img):
+    return f"{DEFAULT_CONTAINER_REGISTRY}/{img}"
+
+
+class TestGetImage:
+    @pytest.fixture
+    def build(self, linux):
+        return Build(linux)
+
+    @pytest.fixture
+    def toolchain_get_image(self, mocker):
+        return mocker.patch("tuxmake.toolchain.Toolchain.get_image")
+
+    def test_image(self, build, toolchain_get_image):
+        toolchain_get_image.return_value = "foobarbaz"
+        assert get_image(build) == q("foobarbaz")
+
+    def test_override_image(self, build, monkeypatch):
+        monkeypatch.setenv("TUXMAKE_IMAGE", "foobar")
+        assert get_image(build) == q("foobar")
+
+    def test_override_image_registry(self, build, monkeypatch, toolchain_get_image):
+        monkeypatch.setenv("TUXMAKE_IMAGE_REGISTRY", "foobar.com")
+        toolchain_get_image.return_value = "myimage"
+        assert get_image(build) == "foobar.com/myimage"
+
+    def test_override_image_tag(self, build, monkeypatch, toolchain_get_image):
+        monkeypatch.setenv("TUXMAKE_IMAGE_TAG", "20201201")
+        toolchain_get_image.return_value = "myimage"
+        assert get_image(build) == q("myimage:20201201")
+
+    def test_override_image_registry_and_tag(
+        self, build, monkeypatch, toolchain_get_image
+    ):
+        monkeypatch.setenv("TUXMAKE_IMAGE_REGISTRY", "foobar.com")
+        monkeypatch.setenv("TUXMAKE_IMAGE_TAG", "20201201")
+        toolchain_get_image.return_value = "myimage"
+        assert get_image(build) == "foobar.com/myimage:20201201"
+
+    def test_override_full_image_name_with_registry(self, build, monkeypatch):
+        monkeypatch.setenv("TUXMAKE_IMAGE", "docker.io/foo/bar")
+        assert get_image(build) == "docker.io/foo/bar"
+
+    def test_detect_dockerhub_registry_on_image_names_with_only_one_component(
+        self, build, monkeypatch
+    ):
+        monkeypatch.setenv("TUXMAKE_IMAGE", "docker.io/foo")
+        assert get_image(build) == "docker.io/foo"
+
+    def test_detect_localhost_registry_on_image_names_with_only_one_component(
+        self, build, monkeypatch
+    ):
+        monkeypatch.setenv("TUXMAKE_IMAGE", "localhost/foo")
+        assert get_image(build) == "localhost/foo"
+
+
+class TestCompression:
+    def test_default_compression(self, linux):
+        build = Build(tree=linux)
+        assert build.compression is default_compression
+        build.run()
+        artifacts = [str(f.name) for f in build.output_dir.glob("*")]
+        assert "modules.tar.xz" in artifacts
+
+    def test_compression_none(self, linux):
+        build = Build(tree=linux, compression_type="none")
+        build.run()
+        artifacts = [str(f.name) for f in build.output_dir.glob("*")]
+        assert "modules.tar" in artifacts
