@@ -1,3 +1,4 @@
+import os
 import re
 import subprocess
 import pytest
@@ -5,15 +6,13 @@ import pytest
 from tuxmake.build import Build
 from tuxmake.exceptions import InvalidRuntimeError
 from tuxmake.exceptions import RuntimePreparationFailed
-from tuxmake.runtime import get_runtime
-from tuxmake.runtime import DEFAULT_CONTAINER_REGISTRY
 from tuxmake.runtime import Runtime
 from tuxmake.runtime import NullRuntime
 from tuxmake.runtime import DockerRuntime
 from tuxmake.runtime import DockerLocalRuntime
 from tuxmake.runtime import PodmanRuntime
 from tuxmake.runtime import PodmanLocalRuntime
-from tuxmake.wrapper import Wrapper
+from tuxmake.runtime import Terminated
 
 
 @pytest.fixture
@@ -22,21 +21,27 @@ def build(linux):
     return b
 
 
+class TestTerminated:
+    def test_basics(self):
+        with pytest.raises(Terminated):
+            Terminated.handle_signal(9, "BOOM")
+
+
 class TestGetRuntime:
     def test_null_runtime(self):
-        assert isinstance(get_runtime(None), NullRuntime)
+        assert isinstance(Runtime.get(None), NullRuntime)
 
     def test_docker_runtime(self):
-        assert isinstance(get_runtime("docker"), DockerRuntime)
+        assert isinstance(Runtime.get("docker"), DockerRuntime)
 
     def test_docker_local_runtime(self):
-        assert isinstance(get_runtime("docker-local"), DockerLocalRuntime)
+        assert isinstance(Runtime.get("docker-local"), DockerLocalRuntime)
 
     def test_invalid_runtime(self):
         with pytest.raises(InvalidRuntimeError):
-            get_runtime("invalid")
+            Runtime.get("invalid")
         with pytest.raises(InvalidRuntimeError):
-            get_runtime("xyz")
+            Runtime.get("xyz")
 
 
 class TestRuntime:
@@ -45,40 +50,58 @@ class TestRuntime:
         with pytest.raises(InvalidRuntimeError):
             Runtime()
 
+    def test_run_cmd_interactive(self, Popen, mocker):
+        get_command_line = mocker.patch(
+            "tuxmake.runtime.Runtime.get_command_line", return_value=["/bin/bash"]
+        )
+        runtime = NullRuntime()
+        runtime.run_cmd(["/bin/bash"], interactive=True)
+        kwargs = Popen.call_args[1]
+        get_command_line.assert_called_with(
+            ["/bin/bash"], interactive=True, offline=True
+        )
+        assert kwargs["stdin"] is None
+        assert kwargs["stdout"] is None
+        assert kwargs["stderr"] is None
+
+    def test_create_output_directory_on_prepare(self, tmp_path):
+        runtime = NullRuntime()
+        runtime.output_dir = tmp_path / "build" / "1"
+        runtime.prepare()
+        assert runtime.output_dir.exists()
+
+    def test_create_output_directory_on_prepare_already_exists(self, tmp_path):
+        runtime = NullRuntime()
+        d = tmp_path / "build" / "1"
+        d.mkdir(parents=True)
+        runtime.output_dir = d
+        runtime.prepare()
+        assert runtime.output_dir.exists()
+
 
 class TestNullRuntime:
     def test_get_command_line(self, build):
         assert NullRuntime().get_command_line(
-            build, ["date"], interactive=False, offline=False
+            ["date"], interactive=False, offline=False
         ) == ["date"]
 
-    def test_prepare_warns_about_versioned_toolchain(self, build, mocker):
-        build.toolchain.version_suffix = "-10"
-        log = mocker.patch("tuxmake.build.Build.log")
-        runtime = NullRuntime()
-        runtime.prepare(build)
-        log.assert_called()
-
     def test_toolchains(self):
         runtime = NullRuntime()
         assert "gcc" in runtime.toolchains
 
 
-@pytest.fixture
-def get_image(mocker):
-    return mocker.patch("tuxmake.toolchain.Toolchain.get_image")
-
-
-def q(img):
-    return f"{DEFAULT_CONTAINER_REGISTRY}/{img}"
-
-
 @pytest.fixture
 def container_id():
     return "0123456789abcdef"
 
 
-class TestContainerRuntime:
+class FakeGetImage:
+    @pytest.fixture(autouse=True)
+    def get_image(self, mocker):
+        return mocker.patch("tuxmake.runtime.Runtime.get_image")
+
+
+class TestContainerRuntime(FakeGetImage):
     @pytest.fixture(autouse=True)
     def spawn_container(self, mocker, container_id):
         return mocker.patch(
@@ -93,48 +116,36 @@ class TestContainerRuntime:
         )
 
 
+class TestGetImage:
+    def test_no_image(self):
+        runtime = NullRuntime()
+        with pytest.raises(Exception):
+            runtime.get_image()
+
+    def test_with_image(self):
+        runtime = NullRuntime()
+        runtime.set_image("myimage")
+        assert runtime.get_image() == "myimage"
+
+
 class TestDockerRuntime(TestContainerRuntime):
-    def test_image(self, build, get_image):
-        get_image.return_value = "foobarbaz"
-        runtime = DockerRuntime()
-        assert runtime.get_image(build) == q("foobarbaz")
+    def test_get_metadata(self, get_image, mocker):
+        get_image.return_value = "tuxmake/theimage"
+        mocker.patch(
+            "subprocess.check_output",
+            return_value=b"tuxmake/theimage@sha256:deadbeef",
+        )
+        metadata = DockerRuntime().get_metadata()
+        assert metadata["image_name"] == "tuxmake/theimage"
+        assert metadata["image_digest"] == "tuxmake/theimage@sha256:deadbeef"
 
-    def test_override_image(self, build, monkeypatch):
-        monkeypatch.setenv("TUXMAKE_IMAGE", "foobar")
-        runtime = DockerRuntime()
-        assert runtime.get_image(build) == q("foobar")
-
-    def test_override_image_registry(self, build, monkeypatch, get_image):
-        monkeypatch.setenv("TUXMAKE_IMAGE_REGISTRY", "foobar.com")
-        get_image.return_value = "myimage"
-        runtime = DockerRuntime()
-        assert runtime.get_image(build) == "foobar.com/myimage"
-
-    def test_override_image_tag(self, build, monkeypatch, get_image):
-        monkeypatch.setenv("TUXMAKE_IMAGE_TAG", "20201201")
-        get_image.return_value = "myimage"
-        runtime = DockerRuntime()
-        assert runtime.get_image(build) == q("myimage:20201201")
-
-    def test_override_image_registry_and_tag(self, build, monkeypatch, get_image):
-        monkeypatch.setenv("TUXMAKE_IMAGE_REGISTRY", "foobar.com")
-        monkeypatch.setenv("TUXMAKE_IMAGE_TAG", "20201201")
-        get_image.return_value = "myimage"
-        runtime = DockerRuntime()
-        assert runtime.get_image(build) == "foobar.com/myimage:20201201"
-
-    def test_override_full_image_name_with_registry(self, build, monkeypatch):
-        monkeypatch.setenv("TUXMAKE_IMAGE", "docker.io/foo/bar")
-        runtime = DockerRuntime()
-        assert runtime.get_image(build) == "docker.io/foo/bar"
-
-    def test_prepare(self, build, get_image, mocker):
+    def test_prepare(self, get_image, mocker):
         get_image.return_value = "myimage"
         check_call = mocker.patch("subprocess.check_call")
-        DockerRuntime().prepare(build)
-        check_call.assert_called_with(["docker", "pull", q("myimage")])
+        DockerRuntime().prepare()
+        check_call.assert_called_with(["docker", "pull", "myimage"])
 
-    def test_prepare_pull_only_once_a_day(self, build, get_image, mocker):
+    def test_prepare_pull_only_once_a_day(self, get_image, mocker):
         get_image.return_value = "myimage"
         check_call = mocker.patch("subprocess.check_call")
         now = 1614000983
@@ -146,69 +157,104 @@ class TestDockerRuntime(TestContainerRuntime):
         )
 
         # first call
-        PodmanRuntime().prepare(build)
+        PodmanRuntime().prepare()
         assert len(check_call.call_args_list) == 1
 
         # after 2 hours, no need to pull
-        PodmanRuntime().prepare(build)
+        PodmanRuntime().prepare()
         assert len(check_call.call_args_list) == 1
 
         # after 2 days, pull again
-        PodmanRuntime().prepare(build)
+        PodmanRuntime().prepare()
         assert len(check_call.call_args_list) == 2
 
-    def test_start_container(self, build, container_id):
+    def test_start_container(self, container_id):
         runtime = DockerRuntime()
-        runtime.start_container(build)
+        runtime.start_container()
         assert runtime.container_id == container_id
 
-    def test_cleanup(self, build, container_id, mocker):
-        check_call = mocker.patch("subprocess.check_call")
+    def test_cleanup(self, container_id, mocker):
+        call = mocker.patch("subprocess.call")
         runtime = DockerRuntime()
-        runtime.start_container(build)
+        runtime.start_container()
         runtime.cleanup()
-        cmd = check_call.call_args[0][0]
+        cmd = call.call_args[0][0]
         assert cmd[0:2] == ["docker", "stop"]
         assert cmd[-1] == container_id
 
-    def test_get_command_line(self, build):
-        cmd = DockerRuntime().get_command_line(build, ["date"], False)
+    def test_cleanup_before_container_exists(self):
+        runtime = DockerRuntime()
+        assert runtime.container_id is None
+        runtime.cleanup()  # if this doesn't crash we are good
+
+    def test_get_command_line(self):
+        cmd = DockerRuntime().get_command_line(["date"], False)
         assert cmd[0:2] == ["docker", "exec"]
         assert cmd[-1] == "date"
 
-    def test_environment(self, build, spawn_container):
-        build.environment = {"FOO": "BAR"}
-        DockerRuntime().start_container(build)
+    def test_environment(self, linux, spawn_container):
+        runtime = DockerRuntime()
+        runtime.environment["FOO"] = "BAR"
+        runtime.start_container()
         cmd = spawn_container.call_args[0][0]
         assert "--env=FOO=BAR" in cmd
 
-    def test_ccache(self, build, home, spawn_container):
-        ccache = Wrapper("ccache")
-        orig_ccache_dir = ccache.environment["CCACHE_DIR"]
-        build.wrapper = ccache
-        DockerRuntime().start_container(build)
+    def test_output_dir(self, linux, tmp_path, spawn_container):
+        runtime = DockerRuntime()
+        runtime.output_dir = tmp_path
+        runtime.start_container()
         cmd = spawn_container.call_args[0][0]
-        assert "--env=CCACHE_DIR=/ccache-dir" in cmd
-        assert f"--volume={orig_ccache_dir}:/ccache-dir" in cmd
+        assert f"--volume={tmp_path}:{tmp_path}" in cmd
 
-    def test_sccache_with_path(self, build, home, spawn_container):
-        sccache_from_host = Wrapper("/opt/bin/sccache")
-        build.wrapper = sccache_from_host
-        DockerRuntime().start_container(build)
+    def test_source_dir(self, linux, tmp_path, spawn_container):
+        runtime = DockerRuntime()
+        runtime.source_dir = tmp_path
+        runtime.start_container()
         cmd = spawn_container.call_args[0][0]
-        assert "--volume=/opt/bin/sccache:/usr/local/bin/sccache" in cmd
+        assert f"--volume={tmp_path}:{tmp_path}" in cmd
 
-    def test_TUXMAKE_DOCKER_RUN(self, build, monkeypatch, spawn_container):
+    def test_volume(self, linux, spawn_container):
+        runtime = DockerRuntime()
+        path = "/path/to/something"
+        runtime.add_volume(path)
+        runtime.start_container()
+        cmd = spawn_container.call_args[0][0]
+        assert f"--volume={path}:{path}" in cmd
+
+    def test_TUXMAKE_DOCKER_RUN(self, monkeypatch, spawn_container):
         monkeypatch.setenv(
             "TUXMAKE_DOCKER_RUN", "--hostname=foobar --env=FOO='bar baz'"
         )
-        DockerRuntime().start_container(build)
+        DockerRuntime().start_container()
         cmd = spawn_container.call_args[0][0]
         assert "--hostname=foobar" in cmd
         assert "--env=FOO=bar baz" in cmd
 
-    def test_interactive(self, build):
-        cmd = DockerRuntime().get_command_line(build, ["bash"], True)
+    def test_set_user(self, spawn_container, mocker):
+        check_call = mocker.patch("subprocess.check_call")
+        runtime = DockerRuntime()
+        runtime.set_user("tuxmake")
+        runtime.start_container()
+        cmd = spawn_container.call_args[0][0]
+        assert "--user=tuxmake" in cmd
+        usermod = check_call.call_args[0][0]
+        uid = str(os.getuid())
+        assert usermod[-4:] == ["usermod", "-u", uid, "tuxmake"]
+
+    def test_set_user_set_group(self, spawn_container, mocker):
+        check_call = mocker.patch("subprocess.check_call")
+        runtime = DockerRuntime()
+        runtime.set_user("tuxmake")
+        runtime.set_group("tuxmake")
+        runtime.start_container()
+        cmd = spawn_container.call_args[0][0]
+        assert "--user=tuxmake:tuxmake" in cmd
+        groupmod = check_call.call_args[0][0]
+        gid = str(os.getgid())
+        assert groupmod[-4:] == ["groupmod", "-g", gid, "tuxmake"]
+
+    def test_interactive(self):
+        cmd = DockerRuntime().get_command_line(["bash"], True)
         assert "--interactive" in cmd
         assert "--tty" in cmd
 
@@ -237,27 +283,27 @@ class TestDockerRuntime(TestContainerRuntime):
         assert str(DockerRuntime()) == "docker"
 
 
-class TestDockerRuntimeSpawnContainer:
-    def test_spawn_container(self, build, mocker, container_id):
+class TestDockerRuntimeSpawnContainer(FakeGetImage):
+    def test_spawn_container(self, mocker, container_id):
         check_output = mocker.patch(
             "subprocess.check_output", return_value=container_id.encode("utf-8")
         )
         runtime = DockerRuntime()
-        runtime.start_container(build)
+        runtime.start_container()
         cmd = check_output.call_args[0][0]
         assert cmd[0:2] == ["docker", "run"]
         assert runtime.container_id == container_id
 
 
-class TestDockerRuntimeOfflineAvailable:
+class TestDockerRuntimeOfflineAvailable(FakeGetImage):
     @pytest.fixture
-    def runtime(self, build, container_id, mocker):
+    def runtime(self, container_id, mocker):
         mocker.patch(
             "tuxmake.runtime.DockerRuntime.spawn_container", return_value=container_id
         )
         mocker.patch("tuxmake.runtime.ContainerRuntime.prepare_image")
         r = DockerRuntime()
-        r.prepare(build)
+        r.prepare()
         return r
 
     def test_offline_available(self, runtime, mocker):
@@ -277,19 +323,19 @@ class TestDockerRuntimeOfflineAvailable:
 
 
 class TestDockerLocalRuntime(TestContainerRuntime):
-    def test_prepare_checks_local_image(self, build, get_image, mocker):
+    def test_prepare_checks_local_image(self, get_image, mocker):
         get_image.return_value = "mylocalimage"
         check_call = mocker.patch("subprocess.check_call")
         runtime = DockerLocalRuntime()
 
-        runtime.prepare(build)
+        runtime.prepare()
         check_call.assert_called_with(
-            ["docker", "image", "inspect", q("mylocalimage")],
+            ["docker", "image", "inspect", "mylocalimage"],
             stdout=subprocess.DEVNULL,
             stderr=subprocess.DEVNULL,
         )
 
-    def test_prepare_image_not_found(self, build, get_image, mocker):
+    def test_prepare_image_not_found(self, get_image, mocker):
         get_image.return_value = "foobar"
         mocker.patch(
             "subprocess.check_call",
@@ -298,8 +344,8 @@ class TestDockerLocalRuntime(TestContainerRuntime):
             ),
         )
         with pytest.raises(RuntimePreparationFailed) as exc:
-            DockerLocalRuntime().prepare(build)
-        assert f"image {q('foobar')} not found locally" in str(exc)
+            DockerLocalRuntime().prepare()
+        assert "image foobar not found locally" in str(exc)
 
     def test_listed_as_supported(self):
         assert "docker-local" in Runtime.supported()
@@ -309,63 +355,63 @@ class TestDockerLocalRuntime(TestContainerRuntime):
 
 
 class TestPodmanRuntime(TestContainerRuntime):
-    def test_prepare(self, build, get_image, mocker):
+    def test_prepare(self, get_image, mocker):
         get_image.return_value = "myimage"
         check_call = mocker.patch("subprocess.check_call")
-        PodmanRuntime().prepare(build)
-        check_call.assert_called_with(["podman", "pull", q("myimage")])
+        PodmanRuntime().prepare()
+        check_call.assert_called_with(["podman", "pull", "myimage"])
 
-    def test_get_command_line(self, build):
-        cmd = PodmanRuntime().get_command_line(build, ["date"], False)
+    def test_get_command_line(self):
+        cmd = PodmanRuntime().get_command_line(["date"], False)
         assert cmd[0:2] == ["podman", "exec"]
         assert cmd[-1] == "date"
 
     def test_listed_as_supported(self):
         assert "podman" in Runtime.supported()
 
-    def test_no_user_option(self, build, spawn_container):
-        PodmanRuntime().start_container(build)
+    def test_no_user_option(self, get_image, spawn_container):
+        PodmanRuntime().start_container()
         cmd = spawn_container.call_args[0][0]
         assert len([c for c in cmd if "--user=" in c]) == 0
 
     def test_str(self):
         assert str(PodmanRuntime()) == "podman"
 
-    def test_TUXMAKE_PODMAN_RUN(self, build, monkeypatch, spawn_container):
+    def test_TUXMAKE_PODMAN_RUN(self, monkeypatch, spawn_container):
         monkeypatch.setenv(
             "TUXMAKE_PODMAN_RUN", "--hostname=foobar --env=FOO='bar baz'"
         )
-        PodmanRuntime().start_container(build)
+        PodmanRuntime().start_container()
         cmd = spawn_container.call_args[0][0]
         assert "--hostname=foobar" in cmd
         assert "--env=FOO=bar baz" in cmd
 
-    def test_selinux_label(self, build, spawn_container):
-        PodmanRuntime().start_container(build)
+    def test_selinux_label(self, get_image, spawn_container):
+        PodmanRuntime().start_container()
         cmd = spawn_container.call_args[0][0]
         volumes = [o for o in cmd if o.startswith("--volume=")]
         assert all([v.endswith(":z") for v in volumes])
 
-    def test_logging_level(self, build, spawn_container):
-        PodmanRuntime().start_container(build)
+    def test_logging_level(self, spawn_container):
+        PodmanRuntime().start_container()
         cmd = spawn_container.call_args[0][0]
         assert "--log-level=ERROR" in cmd
 
 
 class TestPodmanLocalRuntime(TestContainerRuntime):
-    def test_prepare_checks_local_image(self, build, get_image, mocker):
+    def test_prepare_checks_local_image(self, get_image, mocker):
         get_image.return_value = "mylocalimage"
         check_call = mocker.patch("subprocess.check_call")
         runtime = PodmanLocalRuntime()
 
-        runtime.prepare(build)
+        runtime.prepare()
         check_call.assert_called_with(
-            ["podman", "image", "inspect", q("mylocalimage")],
+            ["podman", "image", "inspect", "mylocalimage"],
             stdout=subprocess.DEVNULL,
             stderr=subprocess.DEVNULL,
         )
 
-    def test_prepare_image_not_found(self, build, get_image, mocker):
+    def test_prepare_image_not_found(self, get_image, mocker):
         get_image.return_value = "foobar"
         mocker.patch(
             "subprocess.check_call",
@@ -374,8 +420,8 @@ class TestPodmanLocalRuntime(TestContainerRuntime):
             ),
         )
         with pytest.raises(RuntimePreparationFailed) as exc:
-            PodmanLocalRuntime().prepare(build)
-        assert f"image {q('foobar')} not found locally" in str(exc)
+            PodmanLocalRuntime().prepare()
+        assert "image foobar not found locally" in str(exc)
 
     def test_listed_as_supported(self):
         assert "podman-local" in Runtime.supported()
