@@ -78,6 +78,10 @@ class RuntimePreparationFailed(TuxMakeInfrastructureError):
     msg = "Runtime preparation failed: {name}"
 
 
+class ImageRequired(TuxMakeInfrastructureError):
+    msg = "Image is required, but was not set"
+
+
 class UnsupportedMetadata(TuxMakeInfrastructureError):
     msg = "Unsupported metadata extractor: {name}"
     pass
@@ -88,9 +92,9 @@ class UnsupportedMetadataType(TuxMakeInfrastructureError):
     pass
 
 
-class BuildDirAlreadyExists(TuxMakeUserError):
-    msg = "Build directory {name} already exists. If you are doing parallel builds, you need to specify a build directory explicitly"
-
-
 class EnvironmentCheckFailed(TuxMakeUserError):
     msg = "Environment check failed"
+
+
+class UnsupportedCompression(TuxMakeUserError):
+    msg = "Unsupported compression type: {name}"
